#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "ikna.h"
#include "csv.c"

struct settings settings = {0};
struct lock_fields lock_fields = {0};

static void die(const char *error_string, ...) {
	va_list args;
	va_start(args, error_string);
	vfprintf(stderr, error_string, args);
	exit(1);
}

static void *xmalloc(size_t len) {
	void *ptr = malloc(len);

	if (ptr == NULL) {
		die("ikna: malloc() failed\n");
	}

	return ptr;
}

static void *xrealloc(void *ptr, size_t len) {
	ptr = realloc(ptr, len);

	if (ptr == NULL) {
		die("ikna: realloc() failed\n");
	}

	return ptr;
}

static FILE *xfopen(const char *file_path, const char *mode) {
	FILE *file_ptr = fopen(file_path, mode);

	if (file_ptr == NULL) {
		die("ikna: Could not open file '%s'\n", file_path);
	}

	return file_ptr;
}

static ulong string_to_ulong(const char *s) {
	uint i;
	ulong n;
	n = 0;

	for (i = 0; s[i] >= '0' && s[i] <= '9'; ++i) {
		n = 10 * n + (s[i] - '0');
	}

	return n;
}

/* Keep this here just in case, but comment out to get rid of compiler warning
static long string_to_long(const char *s) {
	uint i;
	long n;
	int sign;

	i = 0;

	if (s[0] == '-') {
		sign = -1;
	} else {
		sign = 1;
	}

	if (s[0] == '-' || s[0] == '+') {
		++i;
	}

	for (n = 0; s[i] >= '0' && s[i] <= '9'; ++i) {
		n = 10 * n + (s[i] - '0');
	}

	return sign * n;
}
*/

/* Does malloc() for face_front and face_back; need to free() it yourself! */
static void get_card_faces(const char *card_path, char **face_front, char **face_back) {
	int ch;
	uint line_len, front_len, back_len, current_card_face, line_buff_size, front_buff_size, back_buff_size;
	char *front, *back, *line_buff;
	FILE *card_file;

	card_file = xfopen(card_path, "r");

	current_card_face = OPTION_FRONT;

	line_buff_size = 128;
	line_buff = xmalloc(sizeof(char) * line_buff_size);

	front_buff_size = line_buff_size;
	back_buff_size = line_buff_size;
	front = xmalloc(sizeof(char) * front_buff_size);
	back = xmalloc(sizeof(char) * back_buff_size);

	front[0] = '\0';
	back[0] = '\0';

	ch = '\0';
	line_len = 0;
	front_len = 0;
	back_len = 0;

	for (;;) {
		/* get entire line into array (realloc if needed) */
		for (;;) {
			ch = getc(card_file);

			if (ch == EOF) {
				break;
			}

			if (line_len >= line_buff_size - 1) {
				line_buff_size *= 2;
				line_buff = xrealloc(line_buff, sizeof(char) * line_buff_size);
			}

			if (front_len >= front_buff_size - 1) {
				front_buff_size *= 2;
				front = xrealloc(front, sizeof(char) * front_buff_size);
			}

			if (back_len >= back_buff_size - 1) {
				back_buff_size *= 2;
				back = xrealloc(back, sizeof(char) * back_buff_size);
			}

			line_buff[line_len] = ch;

			if (line_buff[line_len] == '\n') {
				line_buff[line_len + 1] = '\0';
				line_len = 0;
				break;
			}

			if (current_card_face == OPTION_FRONT) {
				++front_len;
			} else if (current_card_face == OPTION_BACK) {
				++back_len;
			}

			++line_len;
		}

		if (ch == EOF) {
			break;
		}

		/* check if line is ---\n and change current_card_face */
		if (strcmp(line_buff, "---\n") == 0) {
			current_card_face = OPTION_BACK;
			 /* Don't want to print the front/back delimiter */
			continue;
		}

		if (current_card_face == OPTION_FRONT) {
			strcat(front, line_buff);
		} else if (current_card_face == OPTION_BACK) {
			strcat(back, line_buff);
		}
	}

	fclose(card_file);
	free(line_buff);

	*face_front = front;
	*face_back = back;
}

static void new_card(const char *front_string, const char *back_string, uint new_card_id) {
	/* TODO: Currently assumes that there's always at least one entry in the log file for card id 0. Might want to make this more robust */
	char new_card_path[MAX_CARD_ID_BUFFER_SIZE];
	FILE *new_card_file;

	sprintf(new_card_path, "%u", new_card_id);

	new_card_file = xfopen(new_card_path, "w");

	if (front_string == NULL) {
		front_string = "";
	}

	if (back_string == NULL) {
		back_string = "";
	}

	/* TODO: This currently puts a newline when there's NULL string arguments. As a design thing, would it make more sense to have newlines in blank spaces, or nothing? */
	fprintf(new_card_file, "%s\n---\n%s\n", front_string, back_string);

	fclose(new_card_file);

	append_log_file("ikna.log", new_card_id, 'n', time(NULL), settings.default_interval, settings.default_multiplier);

	printf("%s\n", new_card_path);
}

static void print_card(const char *card_path, uint face_to_print) {
	char *front;
	char *back;

	get_card_faces(card_path, &front, &back);

	if (face_to_print == OPTION_FRONT) {
		printf("%s", front);
	} else if (face_to_print == OPTION_BACK) {
		printf("%s", back);
	}

	free(front);
	free(back);
}

static uint read_file_to_char(FILE *file, int *ch, char *buff, uint buff_size, char seek_to) {
	uint count = 0;

	while ((*ch = getc(file)) != seek_to) {
		if (count > buff_size - 1) {
			break;
		}

		if (*ch == EOF) {
			break;
		}

		buff[count] = *ch;
		++count;
	}

	buff[count] = '\0';

	return count;
}

static uint split_string_on_char(char *string, char *split_array[], char split_char, uint max_fields) {
	uint split_count = 0;

	char *p = string;
	uint i = 0;

	split_array[0] = p;
	++split_count;

	while (*p != '\0') {
		if (*p == split_char) {
			if (split_count < max_fields) {
				split_array[split_count] = p + 1;
				string[i] = '\0';
			}

			++split_count;
		}

		++p;
		++i;
	}

	return split_count;
}

static uint parse_card_data(const char *log_path, struct card_data **card_array) {
	time_t today;

	FILE *log_file = xfopen(log_path, "r");

	uint highest_id = 0;

	char line_buff[MAX_LOG_FILE_LINE_BUFFER_SIZE];
	uint line_count = 0;

	uint card_array_size = 128;
	struct card_data *cards = xmalloc(sizeof(struct card_data) * card_array_size);
	uint *id_lines = xmalloc(sizeof(uint) * card_array_size);

	int ch = '\n';

	/* First iteration over file. Get list of lines that have the most recent id's */
	for (;;) {
		uint id, field_count;
		char *line_splits[LOG_FILE_MAX_FIELDS];

		read_file_to_char(log_file, &ch, line_buff, MAX_LOG_FILE_LINE_BUFFER_SIZE, '\n');

		if (ch == EOF) {
			break;
		}

		field_count = split_string_on_char(line_buff, line_splits, ' ', LOG_FILE_MAX_FIELDS);

		if (field_count != LOG_FILE_MAX_FIELDS) {
			break;
		}

		id = string_to_ulong(line_splits[1]);

		if (id >= card_array_size) {
			card_array_size *= 2;
			cards = xrealloc(cards, sizeof(struct card_data) * card_array_size);
			id_lines = xrealloc(id_lines, sizeof(uint) * card_array_size);
		}

		id_lines[id] = line_count;

		if (highest_id < id) {
			highest_id = id;
		}

		++line_count;
	}

	rewind(log_file);
	line_count = 0;
	today = time(NULL);

	/* Go through the file again, this time filling out the card struct for _just_ the most recent lines with the most relevent data */
	for (;;) {
		uint record_id, field_count;
		char *line_splits[LOG_FILE_MAX_FIELDS];

		read_file_to_char(log_file, &ch, line_buff, MAX_LOG_FILE_LINE_BUFFER_SIZE, '\n');

		if (ch == EOF) {
			break;
		}

		field_count = split_string_on_char(line_buff, line_splits, ' ', LOG_FILE_MAX_FIELDS);

		if (field_count != LOG_FILE_MAX_FIELDS) {
			break;
		}

		record_id = string_to_ulong(line_splits[1]);

		if (id_lines[record_id] == line_count) {
			char card_action;
			struct tm record_reviewed = {0};
			double time_diff;
			char *date_field_splits[DATE_MAX_FIELDS];
			int time_diff_over;

			/* 0-imestamp 1-id 2-action 3-date 4-interval 5-mult */
			cards[record_id].id = record_id;

			card_action = *line_splits[2];
			cards[record_id].card_action = card_action;

			split_string_on_char(line_splits[3], date_field_splits, '-', DATE_MAX_FIELDS);

			record_reviewed.tm_year = (string_to_ulong(date_field_splits[0])) - 1900;
			record_reviewed.tm_mon = (string_to_ulong(date_field_splits[1])) - 1;
			record_reviewed.tm_mday = (string_to_ulong(date_field_splits[2]));
			cards[record_id].review_time = mktime(&record_reviewed);

			cards[record_id].interval = string_to_ulong(line_splits[4]);

			cards[record_id].multiplier = atof(line_splits[5]);

			time_diff = difftime(today, cards[record_id].review_time);

			/* Don't want to re-review the card we just studied */
			cards[record_id].ready_for_review = 0;
			time_diff_over = 0;

			if (time_diff > 0) {
				cards[record_id].ready_for_review = 1;
				time_diff_over = 1;
			}

			if (card_action == 'd') {
				cards[record_id].ready_for_review = 0;
			}

			if (card_action == 'u' && time_diff_over) {
				cards[record_id].ready_for_review = 1;
			}
		}

		++line_count;
	}

	free(id_lines);
	fclose(log_file);

	*card_array = cards;
	return highest_id;
}

static void append_log_file(const char *log_path, uint id, char action, time_t card_time, uint interval, float mult) {
	FILE *log_file = xfopen(log_path, "a");
	time_t t;
	struct tm *t_tm, today, card_day;

	log_file = xfopen(log_path, "a");

	t = time(NULL);
	t_tm = localtime(&t);
	today = *t_tm;
	t_tm = localtime(&card_time);
	card_day = *t_tm;

	fprintf(log_file, "%d-%02d-%02dT%02d:%02d:%02dZ %u %c %d-%02d-%02d %u %.2f\n", today.tm_year + 1900, today.tm_mon + 1, today.tm_mday, today.tm_hour, today.tm_min, today.tm_sec, id, action, card_day.tm_year + 1900, card_day.tm_mon + 1, card_day.tm_mday, interval, mult);

	fclose(log_file);
}

static struct lock_fields get_lock_fields(char *lock_file_path) {
	struct lock_fields current_lock_fields = {0};

	FILE *lock_file = xfopen(lock_file_path, "r");

	uint buff_size = MAX_LOCK_FILE_BUFFER_SIZE;
	char *buff = xmalloc(sizeof(char) * buff_size);
	char *field_splits[LOCK_FIELD_MAX];
	int ch;

	char *date_field_splits[DATE_MAX_FIELDS];
	struct tm last_reviewed = {0};

	read_file_to_char(lock_file, &ch, buff, buff_size, '\n');
	split_string_on_char(buff, field_splits, ' ', LOCK_FIELD_MAX);

	current_lock_fields.has_current_card = 1;
	if (field_splits[LOCK_FIELD_CURRENT_CARD][0] == '-') {
		current_lock_fields.has_current_card = 0;
	} else {
		current_lock_fields.current_card_id = string_to_ulong(field_splits[LOCK_FIELD_CURRENT_CARD]);
		current_lock_fields.has_current_card = 1;
	}

	split_string_on_char(field_splits[LOCK_FIELD_LAST_REVIEWED], date_field_splits, '-', DATE_MAX_FIELDS);

	last_reviewed.tm_year = (string_to_ulong(date_field_splits[0])) - 1900;
	last_reviewed.tm_mon = (string_to_ulong(date_field_splits[1])) - 1;
	last_reviewed.tm_mday = (string_to_ulong(date_field_splits[2]));

	current_lock_fields.last_reviewed = mktime(&last_reviewed);
	current_lock_fields.review_count = string_to_ulong(field_splits[LOCK_FIELD_REVIEW_COUNT]);
	current_lock_fields.new_cards_count = string_to_ulong(field_splits[LOCK_FIELD_NEW_CARDS_COUNT]);

	free(buff);
	fclose(lock_file);

	return current_lock_fields;
}

static void set_lock_fields(char *lock_file_path, struct lock_fields lock_fields_to_set) {
	FILE *lock_file = xfopen(lock_file_path, "w");
	struct tm today = *localtime(&lock_fields.last_reviewed);

	char *current_card_string_ptr = xmalloc(sizeof(char) * MAX_CARD_ID_BUFFER_SIZE);
	char *current_card_string;

	if (lock_fields.has_current_card) {
		sprintf(current_card_string_ptr, "%u", lock_fields.current_card_id);
		current_card_string = current_card_string_ptr;
	} else {
		current_card_string = "-";
	}

	fprintf(lock_file, "%s %d-%02d-%02d %u %u\n", current_card_string, today.tm_year + 1900, today.tm_mon + 1, today.tm_mday, lock_fields_to_set.review_count, lock_fields_to_set.new_cards_count);

	fclose(lock_file);
	free(current_card_string_ptr);
}

/* TODO: This function might be able to be factored out completly. It's basically here as legacy from early bad decisions about the lock file. Check out all the things that call to it, to see what I mean */
static uint current_card_id(int *valid_current_card) {
	uint id = 0;

	*valid_current_card = lock_fields.has_current_card;
	if (*valid_current_card) {
		id = lock_fields.current_card_id;
	}

	return id;
}

static void set_next_current_card(struct card_data card_array[], uint highest_id) {
	uint i;
	int reviewing = 0;

	uint *ready_ids = xmalloc(sizeof(uint) * highest_id);
	uint ready_count = 0;

	int at_max_new_cards = 0;
	uint new_cards_today = 0;

	new_cards_today = lock_fields.new_cards_count;

	if (new_cards_today >= settings.max_daily_new_cards) {
		at_max_new_cards = 1;
	}

	for (i = 0; i <= highest_id; ++i) {
		if (card_array[i].ready_for_review) {
			if (at_max_new_cards) {
				if (card_array[i].card_action != 'n') {
					ready_ids[ready_count] = i;
					++ready_count;
					reviewing = 1;
				}
			} else {
				ready_ids[ready_count] = i;
				++ready_count;
				reviewing = 1;
			}
		}
	}

	if (reviewing) {
		uint next;
		srand(time(NULL));
		next = rand() % ready_count;

		lock_fields.current_card_id = ready_ids[next];
		lock_fields.has_current_card = 1;
	} else {
		lock_fields.has_current_card = 0;
	}

	free(ready_ids);
}

static time_t epoch_today_day_start(void) {
	time_t now = time(NULL);

	struct tm today = *localtime(&now);

	today.tm_hour = 0;
	today.tm_min = 0;
	today.tm_sec = 0;

	return mktime(&today);
}

static void reset_review_count_on_new_day(void) {
	double time_diff;

	time_t last_reviewed = lock_fields.last_reviewed;
	time_t today = epoch_today_day_start();

	time_diff = difftime(today, last_reviewed);

	if (time_diff > 0) {
		lock_fields.review_count = 0;
		lock_fields.new_cards_count = 0;
	}
}

static int reviewing_cards_today(struct card_data card_array[], uint highest_id) {
	uint i, reviews_today;
	int cards_ready = 0;

	for (i = 0; i <= highest_id; ++i) {
		if (card_array[i].ready_for_review) {
			cards_ready = 1;
		}
	}

	reviews_today = lock_fields.review_count;

	if (reviews_today >= settings.max_daily_reviews) {
		return 0;
	}

	if (cards_ready) {
		return 1;
	} else {
		return 0;
	}
}

static uint validate_card_id_argument(const char *id_string, uint highest_id) {
	uint i, input_id;

	for (i = 0; *(id_string + i) != '\0'; ++i) {
		char c = *(id_string + i);
		if (!isdigit(c)) {
			die("ikna: Invalid argument '%s'. Must be a positive number\n", id_string);
		}
	}

	/* TODO: what to do if the argument is so large that it just integer overflows? */
	input_id = string_to_ulong(id_string);

	if (input_id > highest_id) {
		die("ikna: Input id '%u' is too large. Highest id in deck is '%u'\n", input_id, highest_id);
	}

	return input_id;
}

static int is_string_valid_int(const char *string) {
	int i;
	int is_valid = 1;
	char c = '\0';

	for (i = 0; *(string + i) != '\0'; ++i) {
		c = *(string + i);
		if (!isdigit(c)) {
			is_valid = 0;
			break;
		}
	}

	return is_valid;
}

static struct settings parse_settings_file(const char *settings_file_path) {
	int is_valid_field = 1;
	struct settings settings = {0};
	FILE *settings_file;
	int ch = '\0';
	char buff[MAX_SETTINGS_FILE_LINE_BUFFER_SIZE];
	char *field_splits[SETTINGS_FILE_MAX_FIELDS];
	uint field_count;

	settings_file = xfopen(settings_file_path, "r");

	for (;;) {
		int check_int = 1;
		read_file_to_char(settings_file, &ch, buff, MAX_SETTINGS_FILE_LINE_BUFFER_SIZE, '\n');

		if (ch == EOF) {
			break;
		}

		field_count = split_string_on_char(buff, field_splits, ' ', SETTINGS_FILE_MAX_FIELDS);

		if (field_count != 2) {
			die("ikna: Setting '%s' can only have two space separated fields\n", field_splits[0]);
		}

		if (strcmp(field_splits[0], "max_daily_reviews") == 0) {
			settings.max_daily_reviews = string_to_ulong(field_splits[1]);
		} else if (strcmp(field_splits[0], "max_daily_new_cards") == 0) {
			settings.max_daily_new_cards = string_to_ulong(field_splits[1]);
		} else if (strcmp(field_splits[0], "default_interval") == 0) {
			settings.default_interval = string_to_ulong(field_splits[1]);
		} else if (strcmp(field_splits[0], "default_multiplier") == 0) {
			settings.default_multiplier = atof(field_splits[1]);

			if (settings.default_multiplier == 0 || field_splits[1][0] == '-') {
				is_valid_field = 0;
			}

			check_int = 0;
		} else {
			die("ikna: Invalid setting '%s'\n", field_splits[0]);
		}

		if (check_int) {
			if (!is_string_valid_int(field_splits[1])) {
				is_valid_field = 0;
			}
		}

		if (!is_valid_field) {
			die("ikna: Invalid field '%s' for setting '%s'\n", field_splits[1], field_splits[0]);
		}
	}

	fclose(settings_file);

	return settings;
}

int main(int argc, char *argv[]) {
	struct card_data *card_list;

	uint option = OPTION_NOTHING;
	int args_max = 2;

	if (argc == 1) {
		printf( "ikna - spaced repetition flash cards\n"
			"\nCommands (ikna <command> <args>):\n"
			" init                     create a new ikna deck in current directory\n"
			" new                      make a new card\n"
			" front/back               print the front/back face of the current card\n"
			"   front/back <id>        pick the card id to print\n"
			" correct/incorrect        answer the current card\n");
		printf( " delete/undelete <id>     delete/restore a card safely\n"
			" status                   print information about current deck\n"
			" list                     print the status of all cards in deck\n"
			" export                   print out current deck in CSV format\n"
			" import                   import cards in CSV format\n"
			"   import/export <delim>  pick a delimiter char for import/export\n"
			" -v/--version             print version information\n");
		printf( "\nSee man page for extended information\n");
		return 0;
	} else {
		/* TODO: There's got to be a better way of testing this without calling strcmp() so many different times... */
		if (strcmp(argv[1], "front") == 0) {
			option = OPTION_FRONT;
			args_max = 3;
		} else if (strcmp(argv[1], "back") == 0) {
			option = OPTION_BACK;
			args_max = 3;
		} else if (strcmp(argv[1], "correct") == 0) {
			option = OPTION_CORRECT;
		} else if (strcmp(argv[1], "incorrect") == 0) {
			option = OPTION_INCORRECT;
		} else if (strcmp(argv[1], "status") == 0) {
			option = OPTION_STATUS;
		} else if (strcmp(argv[1], "list") == 0) {
			option = OPTION_LIST;
		} else if (strcmp(argv[1], "new") == 0) {
			option = OPTION_NEW;
		} else if (strcmp(argv[1], "delete") == 0) {
			option = OPTION_DELETE;
			args_max = 3;
		} else if (strcmp(argv[1], "undelete") == 0) {
			option = OPTION_UNDELETE;
			args_max = 3;
		} else if (strcmp(argv[1], "import") == 0) {
			option = OPTION_IMPORT;
			args_max = 3;
		} else if (strcmp(argv[1], "export") == 0) {
			option = OPTION_EXPORT;
			args_max = 3;
		} else if (strcmp(argv[1], "init") == 0) {
			option = OPTION_INIT;
		} else if (strcmp(argv[1], "-v") == 0 || strcmp(argv[1], "--version") == 0) {
			printf("v0.8\n");
			return 0;
		} else {
			die("ikna: Invalid argument '%s'\n", argv[1]);
		}
	}

	if (argc > args_max) {
		die("ikna: Too many arguments for '%s'\n", argv[1]);
	}

	if (option != OPTION_NOTHING && option != OPTION_INIT) {
		settings = parse_settings_file("ikna.deck");
		lock_fields = get_lock_fields("ikna.lock");
		reset_review_count_on_new_day();
	}

	if (option == OPTION_STATUS) {
		uint i;
		uint highest_id, total_cards_ready_for_review, new_cards_ready_for_review, review_cards_ready_for_review;
		uint current_card, reviewed_today, new_cards_reviewed_today;
		char *current_card_string_ptr = xmalloc(sizeof(char) * MAX_CARD_ID_BUFFER_SIZE);
		char *current_card_string;
		int valid_current_card;

		highest_id = parse_card_data("ikna.log", &card_list);
		valid_current_card = 1;
		current_card_id(&valid_current_card);

		if (!valid_current_card && reviewing_cards_today(card_list, highest_id)) {
			set_next_current_card(card_list, highest_id);
		}

		current_card = lock_fields.current_card_id;
		reviewed_today = lock_fields.review_count;
		new_cards_reviewed_today = lock_fields.new_cards_count;

		total_cards_ready_for_review = 0;
		review_cards_ready_for_review = 0;
		new_cards_ready_for_review = 0;
		for (i = 0; i <= highest_id; ++i) {
			if (card_list[i].ready_for_review == 1) {
				++total_cards_ready_for_review;

				if (card_list[i].card_action == 'n') {
					++new_cards_ready_for_review;
				} else {
					++review_cards_ready_for_review;
				}
			}
		}

		if (lock_fields.has_current_card) {
			sprintf(current_card_string_ptr, "%u", current_card);
			current_card_string = current_card_string_ptr;
		} else {
			current_card_string = "-";
		}

		printf("current_card_id %s\n"
			"cards_reviewed_today %u\n"
			"new_cards_reviewed_today %u\n"
			"max_daily_reviews %u\n"
			"max_daily_new_cards %u\n"
			"highest_card_id %u\n"
			"total_cards_ready_for_review %u\n"
			"review_cards_ready_for_review %u\n"
			"new_cards_ready_for_review %u\n",
			current_card_string, reviewed_today, new_cards_reviewed_today, settings.max_daily_reviews, settings.max_daily_new_cards, highest_id, total_cards_ready_for_review, review_cards_ready_for_review, new_cards_ready_for_review);

		free(current_card_string_ptr);
	}

	if (option == OPTION_LIST) {
		uint i, highest_id;
		struct tm date;
		int valid_current_card;
		int new_card, deleted_card;

		highest_id = parse_card_data("ikna.log", &card_list);
		valid_current_card = 1;
		current_card_id(&valid_current_card);

		if (!valid_current_card && reviewing_cards_today(card_list, highest_id)) {
			set_next_current_card(card_list, highest_id);
		}

		for (i = 0; i <= highest_id; ++i) {
			new_card = 0;
			deleted_card = 0;

			switch (card_list[i].card_action) {
			case 'd':
				deleted_card = 1;
				break;
			case 'n':
				new_card = 1;
				break;
			}

			date = *localtime(&(card_list[i].review_time));
			printf("id:%d ready:%u action:%c deleted:%d new:%d interval:%u multiplier:%.2f date:%d-%02d-%02d\n", card_list[i].id, card_list[i].ready_for_review, card_list[i].card_action, deleted_card, new_card, card_list[i].interval, card_list[i].multiplier, date.tm_year + 1900, date.tm_mon + 1, date.tm_mday);
		}
	}

	if (option == OPTION_FRONT || option == OPTION_BACK) {
		uint highest_id;
		int valid_current_card = 1;
		uint id;

		if (argc == 3) {
			highest_id = parse_card_data("ikna.log", &card_list);
			id = validate_card_id_argument(argv[2], highest_id);
		} else {
			id = current_card_id(&valid_current_card);
		}

		if (!valid_current_card) {
			highest_id = parse_card_data("ikna.log", &card_list);

			if (reviewing_cards_today(card_list, highest_id)) {
				set_next_current_card(card_list, highest_id);
				id = current_card_id(&valid_current_card);
			}
		}

		if (valid_current_card) {
			uint buff_size = MAX_CARD_ID_BUFFER_SIZE;
			char *card_path = xmalloc(sizeof(char) * buff_size);
			sprintf(card_path, "%u", id);

			print_card(card_path, option);
			free(card_path);
		}
	}

	if (option == OPTION_CORRECT || option == OPTION_INCORRECT) {
		int valid_current_card = 1;
		uint id = current_card_id(&valid_current_card);

		char answer_action = 'i';
		time_t log_card_time = time(NULL);
		uint interval = settings.default_interval;
		float multiplier = settings.default_multiplier;

		uint highest_id = parse_card_data("ikna.log", &card_list);

		if (!valid_current_card) {
			if (reviewing_cards_today(card_list, highest_id)) {
				set_next_current_card(card_list, highest_id);
				id = current_card_id(&valid_current_card);
			}
		}

		if (valid_current_card) {
			int look_for_next_card = 1;
			int is_new_card = 0;

			if (option == OPTION_CORRECT) {
				struct tm *t_tm, card_tm;
				uint reviews_today, new_cards_today;

				card_tm.tm_mday += card_list[id].interval;

				t_tm = localtime(&log_card_time);
				card_tm = *t_tm;

				card_tm.tm_mday += card_list[id].interval;
				interval = card_list[id].interval * card_list[id].multiplier;
				answer_action = 'c';

				log_card_time = mktime(&card_tm);

				card_list[id].ready_for_review = 0;

				printf("%u %c %d-%02d-%02d\n", id, answer_action, card_tm.tm_year + 1900, card_tm.tm_mon + 1, card_tm.tm_mday);

				reviews_today = lock_fields.review_count;
				new_cards_today = lock_fields.new_cards_count;

				if (card_list[id].card_action == 'n') {
					++new_cards_today;
					is_new_card = 1;
				}

				if (reviews_today < settings.max_daily_reviews) {
					++reviews_today;
				}

				if (reviews_today >= settings.max_daily_reviews) {
					lock_fields.has_current_card = 0;
					look_for_next_card = 0;
				}

				lock_fields.review_count = reviews_today;

				if (is_new_card) {
					lock_fields.new_cards_count = new_cards_today;
				}
			} else if (option == OPTION_INCORRECT) {
				struct tm *t_tm = localtime(&log_card_time);
				struct tm card_tm = *t_tm;

				printf("%u %c %d-%02d-%02d\n", id, answer_action, card_tm.tm_year + 1900, card_tm.tm_mon + 1, card_tm.tm_mday);
			}

			multiplier = card_list[id].multiplier;

			append_log_file("ikna.log", id, answer_action, log_card_time, interval, multiplier);

			if (look_for_next_card) {
				set_next_current_card(card_list, highest_id);
			}
		}
	}

	if (option == OPTION_NEW) {
		uint highest_id;
		highest_id = parse_card_data("ikna.log", &card_list);

		new_card(NULL, NULL, highest_id + 1);
	}

	if (option == OPTION_DELETE) {
		uint highest_id;
		uint input_id;
		int already_deleted;

		if (argc < 3) {
			die("ikna: 'delete' needs a card id to delete; eg: ikna delete 0\n");
		}

		highest_id = parse_card_data("ikna.log", &card_list);
		input_id = validate_card_id_argument(argv[2], highest_id);

		already_deleted = 0;
		if (card_list[input_id].card_action == 'd') {
			already_deleted = 1;
		}

		if (already_deleted) {
			die("ikna: Card '%u' already deleted. No action taken\n", input_id);
		} else {
			int valid_current_card;
			uint lock_file_id;

			append_log_file("ikna.log", input_id, 'd', card_list[input_id].review_time, card_list[input_id].interval, card_list[input_id].multiplier);
			printf("%u %c\n", input_id, 'd');

			/* Don't want to review the current_card after it's been deleted, now do we? */
			valid_current_card = 1;
			lock_file_id = current_card_id(&valid_current_card);

			if (lock_file_id == input_id && valid_current_card) {
				card_list[input_id].ready_for_review = 0;
				set_next_current_card(card_list, highest_id);
			}
		}
	}

	if (option == OPTION_UNDELETE) {
		uint highest_id;
		uint input_id;
		int is_deleted;

		if (argc < 3) {
			die("ikna: 'undelete' needs a card id to delete; eg: ikna undelete 0\n");
		}

		highest_id = parse_card_data("ikna.log", &card_list);
		input_id = validate_card_id_argument(argv[2], highest_id);

		is_deleted = 1;
		if (card_list[input_id].card_action != 'd') {
			is_deleted = 0;
		}

		if (is_deleted) {
			append_log_file("ikna.log", input_id, 'u', card_list[input_id].review_time, card_list[input_id].interval, card_list[input_id].multiplier);
			printf("%u %c\n", input_id, 'u');
		} else {
			die("ikna: Can't undelete non-deleted card '%u'\n", input_id);
		}
	}

	if (option == OPTION_EXPORT || option == OPTION_IMPORT) {
		uint highest_id = parse_card_data("ikna.log", &card_list);

		char delim = ',';

		free(card_list); /* Just need the highest_id variable */

		/* Custom delimiter in arg */
		if (argc == 3) {
			uint i;
			uint max_len = 1;

			if (argv[2][0] == '\\') {
				if (argv[2][1] != '\0') {
					max_len = 2;

					switch (argv[2][1]) {
					case 't':
						delim = '\t';
						break;
					default:
						die("ikna: Invalid escape char delimiter '%s'\n", argv[2]);
					}
				} else {
					delim = argv[2][0];
				}
			} else {
				delim = argv[2][0];
			}

			for (i = 0; argv[2][i] != '\0'; ++i) {
				if (i >= max_len) {
					die("ikna: Delimiter argument too long. '%s' requires a single character argument like '\\t' or ';', or no arguments (defaults to ',')\n", argv[1]);
				}
			}
		}

		if (option == OPTION_EXPORT) {
			csv_export(highest_id, delim);
		} else if (option == OPTION_IMPORT) {
			csv_import(highest_id, delim);
		}
	}

	if (option == OPTION_INIT) {
		FILE *lock_file, *log_file, *settings_file, *card_0;
		time_t now = time(NULL);

		lock_file = fopen("ikna.lock", "r");
		log_file = fopen("ikna.log", "r");
		settings_file = fopen("ikna.deck", "r");
		if (lock_file != NULL || log_file != NULL || settings_file != NULL) {
			die("ikna: Deck is already initalized. No changes made to current deck.\n");
		}

		lock_file = xfopen("ikna.lock", "w");
		fprintf(lock_file, "0 1990-01-01 0 0\n");
		fclose(lock_file);

		card_0 = xfopen("0", "w");
		fprintf(card_0, "This is the front of a card!\n"
				"---\n"
				"And this is the back!\n"
				"Don't delete this card file. You can edit it or use 'ikna delete 0' instead.\n"
				"Manually deleting a card file will cause a ton of problems.\n"
				"Read the man page ('man ikna') for detailed usage information.\n");
		fclose(card_0);

		settings_file = xfopen("ikna.deck", "w");
		fprintf(settings_file, "max_daily_reviews 50\n"
				"max_daily_new_cards 10\n"
				"default_interval 1\n"
				"default_multiplier 2.00\n");
		fclose(settings_file);

		settings = parse_settings_file("ikna.deck");
		append_log_file("ikna.log", 0, 'n', now, settings.default_interval, settings.default_multiplier);

		printf("Deck successfully initalized.\n"
			"Please read the manual page for how to use ikna.\n");
	}

	if (option != OPTION_NOTHING) {
		lock_fields.last_reviewed = epoch_today_day_start();

		if (option == OPTION_INIT) {
			lock_fields.has_current_card = 1;
		}

		set_lock_fields("ikna.lock", lock_fields);
	}

	return 0;
}
